/*		 ICP-PCI Kernel module free public release
 *
 *		       Copyright (C) 1996-99 GOM mbH.
 *	         Rebenring 33, 38106 Braunschweig, GERMANY
 *		            All rights reserved
 *
 * $Id: test_icpci.c,v 5.1.2.6 1999/07/13 07:39:00 matthias Exp $
 *
 */

/*
 * Defines:
 */

/* #define	DEBUG_ICPCI */

#define FONT_NAME  	"9x15"
#define CREATOR_PROG	"test_icpci"
#define SAVE_NAME	"test_icpci"

/*
 * Includes:
 */

#include        <time.h>
#include        <stdio.h>
#include        <errno.h>
#include        <fcntl.h>
#include        <unistd.h>
#include        <getopt.h>
#include        <stdlib.h>
#include	<signal.h>

#include        <sys/ipc.h>
#include        <sys/shm.h>
#include        <sys/mman.h>
#include	<sys/user.h>
#include	<sys/ioctl.h>
#include	<sys/param.h>
#include	<sys/poll.h>

#include        <X11/Xos.h>
#include        <X11/Xlib.h>
#include        <X11/Xutil.h>
#include        <X11/Xatom.h>

#include        <X11/extensions/XShm.h>

#include	"icpciExt.h"
#include	"amvsReg.h"
#include	"amdigReg.h"
#include	"ampcvReg.h"
#include	"config.h"

/* #include	<tiffio.h> */

/*
 * Typedefs: 
 */

typedef enum {
 READ_MODE,
 MMAP_MODE,
 BMA_MODE
} operation_mode_t;

typedef enum {
 SIGNAL_MODE,
 SELECT_MODE,
 POLL_MODE
} synchronization_mode_t;

typedef enum {
  GRAB_MODE,
  SNAP_MODE
} acquisition_mode_t;

typedef enum {
  NORMAL_MODE,
  TRIGGER_MODE
} snap_mode_t;

/*
 * Devices
 */

const char    *vs_camera_device[] = { "/dev/ic0vs0",
				      "/dev/ic0vs1",
				      "/dev/ic0vs2",
				      "/dev/ic0vs3",
				      "/dev/ic1vs0",
				      "/dev/ic1vs1",
				      "/dev/ic1vs2",
				      "/dev/ic1vs3",
				      "/dev/ic2vs0",
				      "/dev/ic2vs1",
				      "/dev/ic2vs2",
				      "/dev/ic2vs3",
				      "/dev/ic3vs0",
				      "/dev/ic3vs1",
				      "/dev/ic3vs2",
				      "/dev/ic3vs3" };

const char    *pv_camera_device[] = { "/dev/ic0pv0",
				      "/dev/ic0pv1",
				      "/dev/ic0pv2",
				      "/dev/ic0pv3",
				      "/dev/ic1pv0",
				      "/dev/ic1pv1",
				      "/dev/ic1pv2",
				      "/dev/ic1pv3",
				      "/dev/ic2pv0",
				      "/dev/ic2pv1",
				      "/dev/ic2pv2",
				      "/dev/ic2pv3",
				      "/dev/ic3pv0",
				      "/dev/ic3pv1",
				      "/dev/ic3pv2",
				      "/dev/ic3pv3" };

const char    *dg_camera_device[] = { "/dev/ic0dg0",
				      "/dev/ic1dg0",
				      "/dev/ic2dg0",
				      "/dev/ic3dg0" };

const char    *camera_ioctl_device[][ICP_BOARDS_MAX] =
				  {
  				    { }, { }, { },
				    { "/dev/ic0vsI",
				      "/dev/ic1vsI",
				      "/dev/ic2vsI",
				      "/dev/ic3vsI" },
				    { "/dev/ic0dgI",
				      "/dev/ic1dgI",
				      "/dev/ic2dgI",
				      "/dev/ic3dgI" },
				    { }, { }, { }, { }, { }, { },
				    { }, { }, { }, { }, { },
				    { "/dev/ic0pvI",
				      "/dev/ic0pvI",
				      "/dev/ic0pvI",
				      "/dev/ic0pvI" }
				  };

const char    *ident_ioctl_device[] = { "/dev/ic0idI",
					"/dev/ic1idI",
					"/dev/ic2idI",
					"/dev/ic3idI" };
/*
 * Prototypes
 */

void Sigio_Catcher (void);
void Sigalrm_Catcher (void);
void Clean_Up (void);
int  Copy_Image (void);

/* int  Save_Image (char*, void*, int); */

extern Status   XShmQueryExtension (Display *display);

/*
 * Globals:
 */

const char     *def_cnf_file = "default.txt";

extern char    *optarg;
extern int	optind, opterr, optopt;

struct sigaction copy_action;
struct sigaction timeout_action;
struct sigaction ignore_action;

void	       *image_mmaped = NULL;	/* pointer to mmapped frame buffer
					   (on-board or upper memory) */

size_t		mem_size = 0UL;		/* size of on-board/uppen memory */


void	       *image_buffer = NULL;	/* malloc'ed user space buffer for
					   read(2) interface */

operation_mode_t
		operation_mode = READ_MODE;	/* used to define differnt
						   source and destination
						   buffers */

synchronization_mode_t
		synchronization_mode = SIGNAL_MODE;
						/* you can wait for a new
						   frame with select(2) /
						   poll(2) or a signal can
						   be delivert */

acquisition_mode_t
		acquisition_mode = GRAB_MODE;	/* frames can be acquired
						   in grab or snap mode */

snap_mode_t
		snap_mode = NORMAL_MODE;	/* if you acquire frames with
						   snap you can start auto-
						   maticly or per trigger */

int		zoom_mode = 1;		/* new boards support hardware
					   decimation (also called zoom) */
static volatile int
		timeout = FALSE;	/* shared variable between main program
					   and signal handler */

int		cam_dev_fd = -1;	/* camera file descriptor */

int		shmmode = TRUE;		/* use SHM */

Display        *display = NULL;		/* X11 variables */
GC              context = NULL;
XShmSegmentInfo	ShmInfo;

int		image_width = 0,
  		image_height = 0,
  		image_size = 0;

u_short         colortable[256];

int		src_depth;
int		dst_depth;

XImage         *image = NULL;

/*
 * copy image from a 8 or 16 bit src to a 8 or 16 bit destination
 */ 

int
Copy_Image (void) {

  int		line;
  int		pixel;

  u_char       *dst_8 = NULL;
  u_short      *dst_16 = NULL;

  void	       *src = NULL;
  u_char       *src_8 = NULL;
  /* u_short      *src_16 = NULL; */

  sigset_t	local_set;

  /* specify the buffer to copy from */
  switch (operation_mode) {
  case READ_MODE:
    src = image_buffer;
    break;
  case MMAP_MODE:
  case BMA_MODE:
    src = image_mmaped;
    break;
  default:
    return (ERROR);
  }

  /* if we access the buffer in bma mode, no bma operation should de done */
  /* in signal mode we are called from the handler, we are blocked yet */
  if (operation_mode == BMA_MODE && synchronization_mode != SIGNAL_MODE) {
    
    sigemptyset (&local_set);
    sigaddset (&local_set, SIGIO);
    
    if (sigprocmask (SIG_BLOCK, &local_set, NULL) < 0) {
      perror ("Can't block SIGIO");
      return (ERROR);
    }
  }
  
  /* switch over soucre and destination depth's for copy */
  switch (src_depth) {
  case 8:
    switch (dst_depth) {
    case 8:
      src_8 = src;
      dst_8 = (u_char*)image->data;	  
      for (line = 0; line < image_height; line++) {
	for (pixel = 0; pixel < image_width; pixel++, dst_8++, src_8++)
	  *dst_8 = colortable[*src_8];
	dst_8 += image->bytes_per_line - image_width;
      }
      break;
    case 16:
      src_8 = src;
      dst_16 = (u_short*)image->data;	  
      for (line = 0; line < image_height; line++) {
	for (pixel = 0; pixel < image_width; pixel++, dst_16++, src_8++)
	  *dst_16 = colortable[*src_8];
	dst_16 += image->bytes_per_line / 2 - image_width;
      }
      break;
    default:
      return (ERROR);
    }
    break;
  case 10:
  case 12:
  case 16:
    switch (dst_depth) {
    case 8:
      break;
    case 16:
      break;
    default:
      return (ERROR);
    }
    break;
  default:
    return (ERROR);
  }
  
  if (operation_mode == BMA_MODE && synchronization_mode != SIGNAL_MODE) {
    if (sigprocmask (SIG_UNBLOCK, &local_set, NULL) < 0) {
      perror ("Can't unblock SIGIO");
      return FALSE;
    }
  }

  return (OK);
}

#if (0)
/*
 * save 8 or 16 bit image in tiff format
 */

int
Save_Image (char *description, void* src, int number) {

  TIFF	       *tif;
  
  int		line; 

  u_char        *src_8 = NULL;
  /* u_short       *src_16 = NULL; */
  
  char		creator_prog[NAME_MAX];
  char		save_file[NAME_MAX];

  sprintf (save_file, "%s-%d.tif", SAVE_NAME, number);

  if ((tif = TIFFOpen (save_file, "wb")) == NULL) {
    return (ERROR);
  }

  /* specify various tiff format parameters */
  TIFFSetField (tif, TIFFTAG_IMAGEWIDTH, (unsigned long) image_width);
  TIFFSetField (tif, TIFFTAG_IMAGELENGTH, (unsigned long) image_height);

  switch (src_depth) {
  case 8:
    TIFFSetField (tif, TIFFTAG_MINSAMPLEVALUE, 0L);
    TIFFSetField (tif, TIFFTAG_MAXSAMPLEVALUE, 255L);
    TIFFSetField (tif, TIFFTAG_SAMPLESPERPIXEL, sizeof (u_char));
    TIFFSetField (tif, TIFFTAG_BITSPERSAMPLE, 8);
    break;
  case 10:
  case 12:
  case 16:
    break;
  default:
    return (ERROR);
  }

  TIFFSetField (tif, TIFFTAG_ORIENTATION, ORIENTATION_TOPLEFT);
  TIFFSetField (tif, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG);

  TIFFSetField (tif, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_MINISBLACK);
  TIFFSetField (tif, TIFFTAG_COMPRESSION, COMPRESSION_NONE); 

  TIFFSetField (tif, TIFFTAG_ROWSPERSTRIP, 1);

  TIFFSetField (tif, TIFFTAG_IMAGEDESCRIPTION, description);
  
  sprintf (creator_prog, CREATOR_PROG); 

  TIFFSetField(tif, TIFFTAG_SOFTWARE, creator_prog);

  /* save image line by line */
  switch (src_depth) {
  case 8:
    src_8 = src;
    for (line = 0; line < image_height; line++) {
      if (TIFFWriteScanline(tif, src_8, line, 0) != 1) {
	TIFFClose (tif);
	return (ERROR);
      }
      src_8 += image_width;
    }
    break;
  case 10:
  case 12:
  case 16:
    break;
  default:
    return (ERROR);
  }

  TIFFClose (tif);
  
  return (OK);
}
#endif

/*
 * if in mmap or bma mode: completion of acquisition will raise a signal
 */

void
Sigio_Catcher (void) {

  ICP_PRINTF("|");
  
  if (Copy_Image() < 0) {
    Clean_Up ();
  }

  if (sigaction (SIGIO, &ignore_action, NULL) < 0) {
    perror ("Can't set handler to ignore action");
    Clean_Up ();
  }
}

void
Sigalrm_Catcher (void) {

  if (Copy_Image() < 0) {
    Clean_Up ();
  }

  timeout = TRUE;

  if (sigaction (SIGALRM, &ignore_action, NULL) < 0) {
    perror ("Can't set handler to ignore action");
    Clean_Up ();
  }
}

/*
 * close opened files and free allocted buffers
 */

void
Clean_Up (void) {

  if (ShmInfo.shmaddr) {
    XShmDetach (display, &ShmInfo);
    shmdt (ShmInfo.shmaddr);
    shmctl (ShmInfo.shmid, IPC_RMID, 0);
    image->data = NULL;
  }
  if (image != NULL) {
    XDestroyImage (image);
  }
  if (context != NULL) {
    XFreeGC (display, context);
  }
  if (display != NULL) {
    XCloseDisplay (display);
  }

  if (image_mmaped != NULL && mem_size != 0) {
    munmap (image_mmaped, mem_size);
  }

  if (cam_dev_fd != -1) {
    close (cam_dev_fd);
  }

  ICP_PRINTF("\n");

  exit (EXIT_FAILURE);
}

/*
 * main
 */

int
main (int argc, char *argv[]) {

  int		count;
  int		loops;
  int		error;

  int		one = 1;

  /* default command line arguments */
  int		board_arg = 0;		/* default board */
  int		module_arg = ICP_AMVS;	/* default module */
  int		camera_arg = 0;		/* default camera */
  int		grayscales = 64;	/* default number of grayscales */ 

  /* test|read_config variables */
  char		cam_cnf_file[PATH_MAX];
  				/* temporary config file buffer */
  union icp_dev_t
          	cam_cnf_str;	/* camera configuration structure */

  int		brd;
  int		cam;

  int		boards;		/* fill in by test_config */
  int		module[ICP_BOARDS_MAX];
  int		camera[ICP_BOARDS_MAX][ICP_CAMERAS_MAX];
  char		camera_name[MAX_ENTRY_LENGTH];

  float		frames_per_second = 1.0;
 
  /* bma - buffer handling (bma - mode only) */
  int     	buffer;
  int	  	mem_buffer = 0;		/* active mem-buffer */
  int	  	mem_buffers;		/* available mem-buffers */
  int	  	bma_buffer = 0;		/* active bma-buffer */
  int	  	bma_buffers;		/* available bma-buffers */
  off_t		mem_offset[ICP_BUFFERS_MAX];
  				/* mem-buffer offset relative to base */
  off_t		bma_offset[ICP_BUFFERS_MAX];
  				/* bma-buffer offset relative to base */

  /* select handling (select - mode only) */
  fd_set	wait_fdset;		/* set of filedescriptors */
  
  struct timeval
    	  	wait_time;		/* waiting time */

  /* poll handling */
  struct pollfd wait_fd;
  int wait_timeout;

  sigset_t	local_set;

  /* X11 variables */
  int           screen;
  int		planes;
  Colormap      colormap;
  Window        window;
  XColor        color;
  XGCValues     values;
  Font          timeFont;
  char          windowname[NAME_MAX];

  /* Frames per second evaluation */
  u_char	       *info_ptr = NULL;

  struct icp_info_t
    		*image_info;

  struct timeval
		updatetime,
		starttime,
		frametime[ICP_SNAPS_MAX];

  struct timezone
		dummy_1;

  struct itimerval
    		wait_itimer;

  time_t	counttime = 0,
                xTime = 0;

  float         fps = 0.0;

  char          sfps[64],
                sTime[9];

#if (0)
  char		display_name[NAME_MAX],
    		host_name[NAME_MAX],
    		puffer0[NAME_MAX], puffer1[NAME_MAX], puffer2[NAME_MAX],
    		puffer3[NAME_MAX], puffer4[NAME_MAX], puffer5[NAME_MAX];
#endif

  strcpy (cam_cnf_file, def_cnf_file);

  ShmInfo.shmaddr = NULL;
  
  signal (SIGINT, (void*)Clean_Up); /* catch SIGINT */  

  /* command line parsing -------------------------------------------------- */

  while (1) {
    if ((count = getopt (argc, argv, "f:b:m:n:c:g:o:p:s:artu:z:")) == -1) {
      break;
    }
    switch (count) {
    case 'f':
	if (!optarg) {
	  printf ("No or wrong config file given, using 'default.txt'.\n");
	} else {
	  strcpy (cam_cnf_file, optarg);
	}
      break;
    case 'b':
      if (!optarg || (board_arg = atoi (optarg)) < 0 ||
	  board_arg >= ICP_BOARDS_MAX) {
	printf ("No or wrong board selected, using board 0.\n");
	board_arg = 0;
      }
      break;
    case 'm':
      if (!optarg) {
	printf ("No module selected, using AM-VS.\n");
	module_arg = ICP_AMVS;
      } else {
	if (!strcmp (optarg, "vs")) {
	  module_arg = ICP_AMVS;
	  break;
	}
	if (!strcmp (optarg, "dg")) {
	  module_arg = ICP_AMDG;
	  break;
	}
	if (!strcmp (optarg, "pv")) {
	  module_arg = ICP_AMPV;
	  break;
	}
	printf ("Wrong module selected, using AM-VS.\n");
	module_arg = ICP_AMVS;
      }
      break;
    case 'c':
      if (!optarg || (camera_arg = atoi (optarg)) < 0 || camera_arg > 3) {
	printf ("No or wrong camera given, using camera 0.\n");
	camera_arg = 0;
      }
      break;
    case 'g':
      if (optarg) {
	grayscales = atoi (optarg);
	switch (grayscales) {
	case 8:
	case 16:
	case 32:
	case 64:
	case 128:
	case 256:
	  break;
	default:
	  printf ("No or wrong number of grayscales given, using 64.\n");
	  grayscales = 64;
	}
      }
      break;
    case 'n':
      if (optarg) {
	mem_buffer = atoi (optarg);
	if (mem_buffer < 0 || mem_buffer > ICP_BUFFERS_MAX) {
	  printf ("No or wrong mem-buffer given, using default (0).\n");
	  mem_buffer = 0;
	}
      }
      break;
    case 'o':
      if (!strcmp (optarg, "read")) {
	operation_mode = READ_MODE;
	break;
      }
      if (!strcmp (optarg, "mmap")) {
	operation_mode = MMAP_MODE;
	break;
      }
      if (!strcmp (optarg, "bma")) {
	operation_mode = BMA_MODE;
	break;
      }
      printf ("No or wrong operation mode given, using read interface\n");
      operation_mode = READ_MODE;
      break;
    case 'p':
      if (optarg) {
	frames_per_second = atof (optarg);
	if (frames_per_second < ICP_FPS_MIN ||
	    frames_per_second > ICP_FPS_MAX) {
	  printf ("No or wrong fps value given, using default (1.0).\n");
	  frames_per_second = 1.0;
	}
      }
    case 's':
      if (!strcmp (optarg, "signal")) {
	synchronization_mode = SIGNAL_MODE;
	break;
      }
      if (!strcmp (optarg, "select")) {
	synchronization_mode = SELECT_MODE;
	break;
      }
      if (!strcmp (optarg, "poll")) {
	synchronization_mode = POLL_MODE;
	break;
      }
      printf ("No or wrong operation mode given, using read interface\n");
      operation_mode = READ_MODE;
      break;
    case 'a':
      acquisition_mode = SNAP_MODE;
      break;
    case 't':
      snap_mode = TRIGGER_MODE;
      break;
    case 'r':
      shmmode = FALSE;
      break;
    case 'u':
      if (optarg) {
	bma_buffer = atoi (optarg);
	if (bma_buffer < 0 || bma_buffer > ICP_BUFFERS_MAX) {
	  printf ("No or wrong bma-buffer given, using default (0).\n");
	  bma_buffer = 0;
	}
      }
      break;
    case 'z':
      if (optarg) {
	zoom_mode = atoi (optarg);
	switch (zoom_mode)
	  {
	  case 1:
	  case 2:
	  case 4:
	  case 8:
	    break;
	  default:
	    printf ("Wrong zoom factor selected, using no zoom (1).\n");
	    zoom_mode = 1;
	  }
      }
      break;
    case '?':
    default:
      printf ("Usage: test_icpci -f configuration file (default.txt)\n");
      printf ("                  -b board number (0) [0|1|2|3]\n");
      printf ("                  -m module type (vs) [fa|c1|vs|dg|pv]\n");
      printf ("                  -c camera number (0) [0|1|2|3]\n");
      printf ("                  -g grayscales (64)[8|16|32|64|128|256]\n");
      printf ("                  -n active on-borad memory buffer (0)\n");
      printf ("                     (only in mmap-mode valid)\n");
      printf ("                  -o operation mode (read)[read|mmap|bma]\n");
      printf ("                  -p frames per second (%.1f)[%.2f-%.1f]\n",
	      frames_per_second, ICP_FPS_MIN, ICP_FPS_MAX);
      printf ("                  -s synchronization mode (signal)\n");
      printf ("                     [signa|select|poll]\n");
      printf ("                  -a use snap instead of grab (off)\n");
      printf ("                  -r don't use XSharedMemoryExtension (on)\n");
      printf ("                  -t wait for trigger to snap image (off)\n");
      printf ("                     (requires also option \"-a\")\n");
      printf ("                  -u active upper memory buffer (0)\n");
      printf ("                     (only in bma-mode valid)\n");
      printf ("                  -z zoom mode selection (1) [1|2|4|8]\n");
      exit (EXIT_FAILURE);
    }
  }

  if (module_arg == ICP_AMDG) {
    if (camera_arg != 0) {
      printf ("Wrong camera given, using camera 0.\n");
      camera_arg = 0;
    }
  }

  if (operation_mode == READ_MODE && acquisition_mode == SNAP_MODE) {
    fprintf (stderr, "Option -a has no effect in read mode.\n");
  }

  if (operation_mode != MMAP_MODE && mem_buffer != 0) {
    fprintf (stderr, "Option -n has no effect if not in mmap mode.\n");
  }

  if (operation_mode != BMA_MODE && bma_buffer != 0) {
    fprintf (stderr, "Option -u has no effect if not in bma mode.\n");
  }

  /* open frame grabber devices ------------------------------------------- */

  switch (operation_mode) {
  case READ_MODE:
    switch (module_arg) {
    case ICP_AMVS:
      if ((cam_dev_fd = open (vs_camera_device[board_arg * 4 + camera_arg],
			      O_RDONLY)) < 0) {
	perror ("open camera device");
	Clean_Up ();
      }
      break;
    case ICP_AMDG:
      if ((cam_dev_fd = open (dg_camera_device[board_arg],
			      O_RDONLY)) < 0) {
	perror ("open camera device");
	Clean_Up ();
      }
      break;
    case ICP_AMPV:
      if ((cam_dev_fd = open (pv_camera_device[board_arg * 4 + camera_arg],
			      O_RDONLY)) < 0) {
	perror ("open camera device");
	Clean_Up ();
      }
      break;
    default:
      perror ("wrong module specified");
      Clean_Up ();
    }
    break;

  case MMAP_MODE:
    if ((cam_dev_fd = open (camera_ioctl_device[module_arg][board_arg],
			    O_RDONLY)) < 0) {
      perror ("open camera device");
      Clean_Up ();
    }
    if (ioctl (cam_dev_fd, GIOC_GE_SET_CAMERA, &camera_arg) < 0) {
      perror ("GIOC_GE_SET_CAMERA");
      Clean_Up ();
    }
    break;

  case BMA_MODE:
    if ((cam_dev_fd = open (ident_ioctl_device[board_arg],
			    O_RDONLY)) < 0) {
      perror ("open camera device");
      Clean_Up ();
    }
    if (ioctl (cam_dev_fd, GIOC_GE_SET_CAMERA, &camera_arg) < 0) {
      perror ("GIOC_GE_SET_CAMERA");
      Clean_Up ();
    }
    if (ioctl (cam_dev_fd, GIOC_GE_GET_MODULE, &module_arg) < 0) {
      perror ("GIOC_GE_GET_MODULE");
      Clean_Up ();
    }
    break;
  default:
    perror ("wrong operation mode specified");
    Clean_Up ();
  }
  
  /* test, read and set configuration -------------------------------------- */

  if (ioctl (cam_dev_fd, GIOC_GE_GET_CAMCNF, &cam_cnf_str) < 0) {
    perror ("GIOC_GE_GET_CAMCNF");
    Clean_Up ();
  }

  if ((error = test_config (cam_cnf_file, &boards, module,
			    camera)) < 0) {
    switch (error) {
    case -EICP_FILE:
      fprintf (stderr, "File not found (file: %s)!\n", cam_cnf_file); 
      break;
    default:
      fprintf (stderr, "Error not specified!\n"); 
    }
    Clean_Up ();
  }

  fprintf (stderr, "Config File supports:\n");
  fprintf (stderr, "\t%d board(s)\n", boards);

  for (brd = 0; brd < ICP_BOARDS_MAX; brd++) {
    switch (module[brd]) {
    case ICP_AMVS:
      fprintf (stderr, "\tboard %d: module %s\n", brd, "AM-VS");
      break;
    case ICP_AMDG:
      fprintf (stderr, "\tboard %d: module %s\n", brd, "AM-DIG");
      break;
    case ICP_AMPV:
      fprintf (stderr, "\tboard %d: module %s\n", brd, "AM-PV");
      break;
    default:
      fprintf (stderr, "\tboard %d: module %s\n", brd, "not detected");
    }
    for (cam = 0; cam < ICP_CAMERAS_MAX; cam++) {
      fprintf (stderr, "\tboard %d: camera %d %s\n", brd, cam,
	       camera[brd * ICP_BOARDS_MAX + cam] ?
	       "supported" : "not supported");
    }
  }
  
  if ((error = read_config (cam_cnf_file, &cam_cnf_str,
			    board_arg, module_arg, camera_arg,
			    camera_name)) < 0) {
    switch (error) {
    case -EICP_FILE:
      fprintf (stderr, "File not found (file: %s)!\n", cam_cnf_file); 
      break;
    case -EICP_BOARD:
      fprintf (stderr, "No board marker found (board: %d)!\n", board_arg); 
      break;
    case -EICP_MODULE:
      fprintf (stderr, "No module marker found (module: %d)!\n", module_arg); 
      break;
    case -EICP_CAMERA:
      fprintf (stderr, "No camera marker found (camera: %d)!\n", camera_arg); 
      break;
    case -EICP_OVERFLOW:
      fprintf (stderr, "Table overflow (should not happen)!\n"); 
      break;
    case -EICP_IDENTIFIER:
      fprintf (stderr, "Wrong identifier found!\n"); 
      break;
    case -EICP_ENUM_PARAMETER:
      fprintf (stderr, "Wrong enumerate parameter found!\n"); 
      break;
    case -EICP_NUM_PARAMETER:
      fprintf (stderr, "Wrong digit parameter found!\n"); 
      break;
    default:
      fprintf (stderr, "Error not specified!\n"); 
    }
    Clean_Up ();
  }

  fprintf (stderr, "CameraName: %s.\n", camera_name);

  if (ioctl (cam_dev_fd, GIOC_GE_SET_CAMCNF, &cam_cnf_str) < 0) {
    perror ("GIOC_GE_SET_CAMCNF");
    Clean_Up ();
    exit (EXIT_FAILURE);
  }

  /* only 8 bit for now */
  if (module_arg == ICP_AMDG) {
    if (ioctl (cam_dev_fd, GIOC_DG_SET_NORM_TO8, NULL) < 0) {
      perror ("GIOC_DG_SET_NORM_TO8");
      Clean_Up ();
    }
  }

  /* set frame rate, if needed */
  if (ioctl (cam_dev_fd, GIOC_GE_SET_FPS, &frames_per_second) < 0) {
    perror ("GOIC_GE_SET_FPS");
    Clean_Up ();
  }

  /* get image dimensions ------------------------------------------------- */

  if (operation_mode != BMA_MODE)
    {
      /* work on mem buffers */
      if (ioctl (cam_dev_fd, GIOC_GE_GET_MEMWIDTH, &image_width) < 0) {
	perror ("GIOC_GE_GET_MEMWIDTH");
	Clean_Up ();
      }
      
      if (ioctl (cam_dev_fd, GIOC_GE_GET_MEMHEIGHT, &image_height) < 0) {
	perror ("GIOC_GE_GET_MEMHEIGHT");
	Clean_Up ();
      }
      
      if (ioctl (cam_dev_fd, GIOC_GE_GET_MEMBUFSIZE, &image_size) < 0) {
	perror ("GIOC_GE_GET_MEMBUFSIZE");
	Clean_Up ();
      }
    }
  else
    {
      /* set decimation (zoom) factor */
      if (ioctl (cam_dev_fd, GIOC_CR_SET_HVDEC, &zoom_mode) < 0) {
	perror ("GOIC_CR_SET_HVDEC");
	Clean_Up ();
      }
      /* work on bma buffers */
      if (ioctl (cam_dev_fd, GIOC_GE_GET_BMAWIDTH, &image_width) < 0) {
	perror ("GIOC_GE_GET_BMAWIDTH");
	Clean_Up ();
      }
      
      if (ioctl (cam_dev_fd, GIOC_GE_GET_BMAHEIGHT, &image_height) < 0) {
	perror ("GIOC_GE_GET_BMAHEIGHT");
	Clean_Up ();
      }
      
      if (ioctl (cam_dev_fd, GIOC_GE_GET_BMABUFSIZE, &image_size) < 0) {
	perror ("GIOC_GE_GET_BMABUFSIZE");
	Clean_Up ();
      }
    }

  src_depth = image_size / (image_width * image_height) * 8;
  ICP_PRINTF("source depth: %d\n", src_depth);

  /* initialize signal interface ------------------------------------------ */ 

  ignore_action.sa_handler = SIG_IGN;
  sigemptyset (&ignore_action.sa_mask);
  ignore_action.sa_flags = 0;

  copy_action.sa_handler = (void*)Sigio_Catcher;
  sigemptyset (&copy_action.sa_mask);
  ignore_action.sa_flags = SA_ONESHOT;

  timeout_action.sa_handler = (void*)Sigalrm_Catcher;
  sigemptyset (&timeout_action.sa_mask);
  ignore_action.sa_flags = SA_ONESHOT;
      
  /* initialize to given operation modes ---------------------------------- */ 

  switch (operation_mode) {
  case READ_MODE:
    if ((image_buffer = (void*)malloc (image_size)) == NULL) {
      printf ("virtual memory exhausted.\n");
      Clean_Up ();
    }
    break;

  case MMAP_MODE:
    if (ioctl (cam_dev_fd, GIOC_GE_GET_MEMBUFFERS, &mem_buffers) < 0) {
      perror ("GIOC_GE_GET_MEMBUFFERS");
      Clean_Up ();
    }
    fprintf (stderr, "No. MEM(on-board)-Buffers: %d.\n", mem_buffers);
    if (mem_buffer >= mem_buffers) {
      printf ("Wrong mem-buffer number given, using buffer 0.\n");
      mem_buffer = 0;
    }

    for (buffer = mem_buffers; buffer-- > 0;) {
      if (ioctl (cam_dev_fd, GIOC_GE_SET_MEMBUF, &buffer) < 0) {
	perror ("GIOC_GE_SET_MEMBUF");
	Clean_Up ();
      }
      if (ioctl (cam_dev_fd, GIOC_GE_GET_MEMBUFADDR,
		 &mem_offset[buffer]) < 0) {
	perror ("GIOC_GE_GET_MEMADDR");
	Clean_Up ();
      }
    }
    if (ioctl (cam_dev_fd, GIOC_GE_GET_MEMMAPSIZE, &mem_size) < 0) {
      perror ("GIOC_GE_GET_MEMSIZE");
      Clean_Up ();
    }
    if ((image_mmaped = (void*)mmap (0, mem_size, PROT_READ, MAP_SHARED,
				     cam_dev_fd, 0)) == (void*)-1) {
      perror ("Can't map camera device");
      Clean_Up ();
    }
    image_mmaped += mem_offset[mem_buffer]; 
    if (ioctl (cam_dev_fd, GIOC_GE_SET_MEMBUF, &mem_buffer) < 0) {
      perror ("GIOC_GE_SET_MEMBUF");
      Clean_Up ();
    }
    break;

  case BMA_MODE:
    if (ioctl (cam_dev_fd, GIOC_GE_GET_BMABUFFERS, &bma_buffers) < 0) {
      perror ("GIOC_GE_GET_BMABUFFERS");
      Clean_Up ();
    }
    fprintf (stderr, "No. BMA(upper-memory)-Buffers: %d.\n", bma_buffers);
    if (bma_buffer >= bma_buffers) {
      printf ("Wrong bma-buffer number given, using buffer 0.\n");
      bma_buffer = 0;
    }

    for (buffer = bma_buffers; buffer-- > 0;) {
      if (ioctl (cam_dev_fd, GIOC_GE_SET_BMABUF, &buffer) < 0) {
	perror ("GIOC_GE_SET_BMABUF");
	Clean_Up ();
      }
      if (ioctl (cam_dev_fd, GIOC_GE_GET_BMABUFADDR,
		 &bma_offset[buffer]) < 0) {
	perror ("GIOC_GE_GET_BMAADDR");
	Clean_Up ();
      }
    }
    if (ioctl (cam_dev_fd, GIOC_GE_GET_BMAMAPSIZE, &mem_size) < 0) {
      perror ("GIOC_GE_GET_BMASIZE");
      Clean_Up ();
    }
    if ((image_mmaped = (void*)mmap (0, mem_size, PROT_READ, MAP_SHARED,
				     cam_dev_fd, 0)) == (void*)-1) {
      perror ("Can't map camera device");
      Clean_Up ();
    }
    image_mmaped += bma_offset[bma_buffer]; 
    if (ioctl (cam_dev_fd, GIOC_GE_SET_BMABUF, &bma_buffer) < 0) {
      perror ("GIOC_GE_SET_BMABUF");
      Clean_Up ();
    }
    break;

  default:
    perror ("wrong operation mode specified");
    Clean_Up ();
  }
  
  if (operation_mode != READ_MODE) {
    
    if (snap_mode == TRIGGER_MODE) {
      if (ioctl (cam_dev_fd, GIOC_CR_SET_TRIGGER, NULL) < 0) {
	perror ("GIOC_CR_SET_TRIGGER");
	Clean_Up ();
      }
    } else {
      if (ioctl (cam_dev_fd, GIOC_CR_SET_NORMAL, NULL) < 0) {
	perror ("GIOC_CR_SET_NORMAL");
	Clean_Up ();
      }
    }

    if (synchronization_mode == SIGNAL_MODE) {
      if (sigaction (SIGIO, &ignore_action, NULL) < 0) {
	perror ("Can't set handler to ignore action");
	Clean_Up ();
      }
      if (fcntl (cam_dev_fd, F_SETOWN, getpid()) < 0) {
	perror ("F_SETOWN");
	Clean_Up ();
      }
      if (fcntl (cam_dev_fd, F_SETFL,
		 fcntl (cam_dev_fd, F_GETFL, NULL) | FASYNC) < 0) {
	perror ("F_SETFL FASYNC");
	Clean_Up ();
      }
    } else {
      /* if (acquisition_mode != SNAP_MODE) { */
        if (fcntl (cam_dev_fd, F_SETFL, 
	  	   fcntl (cam_dev_fd, F_GETFL, NULL) | O_NONBLOCK) < 0) {
	  perror ("F_SETFL O_NONBLOCK");
	  Clean_Up ();
        }
      /* } */
    }
    if (acquisition_mode != SNAP_MODE) {
      if (ioctl (cam_dev_fd, GIOC_CR_SET_GRAB, NULL) < 0) {
	if (errno != ENODATA) {
	  perror ("GIOC_CR_SET_GRAB");
	  Clean_Up ();
	}
      }
    }
  }
    
  /* setup X11 ------------------------------------------------------------- */

  if ((display = XOpenDisplay (NULL)) == NULL) {
      printf ("Can't open display.\n");
      Clean_Up ();
    }

  screen = DefaultScreen (display);
  planes = DisplayPlanes (display, screen);
  colormap = DefaultColormap (display, screen);
  
  if ((window = XCreateSimpleWindow (display, RootWindow (display, screen),
				     0, 0, image_width, image_height, 4, 
				     BlackPixel (display, screen),
				     WhitePixel (display, screen))) == 0) {
    printf ("Can't create window.\n");
    Clean_Up ();
  }

  for (count = 0; count < grayscales; count++) {
    color.pad = 0;
    color.red = color.green = color.blue = (65536 / grayscales) * count;
    if (XAllocColor (display, colormap, &color) == 0) {
      printf ("Can't allocate more colors.\n");
      Clean_Up ();
    }
    switch (grayscales) {
    case 256:
      colortable[count] = color.pixel;
      break;
    case 128:
      colortable[count * 2] = color.pixel;
      colortable[count * 2 + 1] = color.pixel;
      break;
    case 64:
      for (loops = 0; loops < 4; loops++ )
	colortable[count * 4 + loops] = color.pixel;
      break;
    case 32:
      for (loops = 0; loops < 8; loops++ )
	colortable[count * 8 + loops] = color.pixel;
      break;
    case 16:
      for (loops = 0; loops < 16; loops++ )
	colortable[count * 16 + loops] = color.pixel;
      break;
    case 8:
      for (loops = 0; loops < 32; loops++ )
	colortable[count * 32 + loops] = color.pixel;
      break;
    }
  }

  if ((context = XCreateGC (display, window, 0, &values)) == NULL) {
      printf ("Can't create graphics context.\n");
      Clean_Up ();
    }

  XSetForeground (display, context, WhitePixel(display, screen));
  XSetBackground (display, context, BlackPixel(display, screen));

  if ((timeFont = XLoadFont (display, FONT_NAME)) == 0) {
      printf ("Can't load font.\n");
      Clean_Up ();
    }

  XSetFont (display, context, timeFont);

#if (0)
  /* this is a big hack - how to do it right? */
  strcpy (display_name, getenv ("DISPLAY"));

  host_name[0] = '\0';
  gethostname (host_name, NAME_MAX);
  host_name[NAME_MAX - 1] = '\0';
  sprintf (puffer0, "unix:0");
  sprintf (puffer1, "unix:0.0");
  sprintf (puffer2, "localhost:0");
  sprintf (puffer3, "localhost:0.0");
  sprintf (puffer4, "%s:0", host_name);
  sprintf (puffer5, "%s:0.0", host_name);
  
  fprintf (stderr, "%s:0", host_name);

  if (display_name[0] == ':'
      || strncmp (display_name, puffer0, strlen (puffer0)) == 0
      || strncmp (display_name, puffer1, strlen (puffer1)) == 0
      || strncmp (display_name, puffer2, strlen (puffer2)) == 0
      || strncmp (display_name, puffer3, strlen (puffer3)) == 0
      || strncmp (display_name, puffer4, strlen (puffer4)) == 0
      || strncmp (display_name, puffer4, strlen (puffer5)) == 0) {
    shmmode = XShmQueryExtension (display);
  } else {
    shmmode = FALSE;
  }
#endif

  if (shmmode) {
    shmmode = XShmQueryExtension (display);
    if (!shmmode) {
      printf ("XSharedMemoryExtension not available!\n");
      Clean_Up ();
    }
    image = XShmCreateImage (display, DefaultVisual (display, screen),
			     planes, ZPixmap, NULL, &ShmInfo,
			     image_width, image_height);

    ShmInfo.shmid = shmget (IPC_PRIVATE, image_size * planes / 8,
			    IPC_CREAT | 0777);
    image->data = ShmInfo.shmaddr = shmat (ShmInfo.shmid, 0, 0);

    if (!XShmAttach (display, &ShmInfo)) {
      shmdt (ShmInfo.shmaddr);
      shmctl (ShmInfo.shmid, IPC_RMID, 0);
      printf ("Can't attach shared memory");
      Clean_Up ();
    }
  } else {
    image = XCreateImage (display, DefaultVisual (display, screen),
			  planes, ZPixmap, 0, NULL,
			  image_width, image_height, 8, image_width);
    
    image->data = malloc (image_size);
  }

  XMapWindow (display, window);
  
  XStoreName (display, window, "Live Video");
  
  XSync (display, TRUE);  

  dst_depth = planes;
  ICP_PRINTF("destination depth: %d\n", dst_depth);

  /* start acquisition loop ------------------------------------------------ */
  
  gettimeofday (&starttime, &dummy_1);

  for (loops = 1; ; loops++) {

    /* Frames per second */

    if (!(loops % 10)) {
      gettimeofday (&updatetime, &dummy_1);
      fps = updatetime.tv_sec * 1000000 + updatetime.tv_usec - counttime;
      counttime = updatetime.tv_sec * 1000000 + updatetime.tv_usec;
      fps = 10000000 / fps;
      sprintf (sfps, "%3.1f", fps);
      strcpy (windowname, "Frames/sec : ");
      strcat (windowname, sfps); 
      XStoreName (display, window, windowname);
    }

    /* get next image */

    if (snap_mode == TRIGGER_MODE) {
      if (ioctl (cam_dev_fd, GIOC_CR_SET_TRIGGER, NULL) < 0) {
	perror ("GIOC_CR_SET_TRIGGER");
	Clean_Up ();
      }
    }

    errno = EAGAIN; /* on success errno is not set */
    switch (operation_mode) {
    case READ_MODE:
      do {
	if (read (cam_dev_fd, image_buffer, image_size) < 0) {
	  if (errno != EINTR) {
	    if (errno == ENODATA) {
	      strcpy (windowname, "*** No Sync ***");
	      XStoreName (display, window, windowname);
	      break;
	    } else {
	      perror ("read image");
	      Clean_Up ();
	    }
	  }
	} else {
	  ICP_PRINTF("\\");
	  break;
	}
      } while (errno != EINTR);
      if (Copy_Image() < 0) {
	Clean_Up ();
	exit (EXIT_FAILURE);
      }
      break;
    case MMAP_MODE:
    case BMA_MODE:
      if (synchronization_mode == SIGNAL_MODE /* &&
	  acquisition_mode != SNAP_MODE */) {
	if (sigaction (SIGIO, &copy_action, NULL) < 0) {
	  perror ("Can't install signal handler");
	  Clean_Up ();
	}
	timeout = FALSE;
	if (sigaction (SIGALRM, &timeout_action, NULL) < 0) {
	  perror ("Can't install signal handler");
	  Clean_Up ();
	}
	timeout = FALSE;
	wait_itimer.it_interval.tv_sec = 0; /* only once */
	wait_itimer.it_interval.tv_usec = 0;
	wait_itimer.it_value.tv_sec = 1;
	wait_itimer.it_value.tv_usec = 0;
	if (setitimer (ITIMER_REAL, &wait_itimer, NULL) != 0) {
	  perror ("Can't install intervall timer");
	  Clean_Up ();
	}
      }
      if (acquisition_mode == SNAP_MODE) {
	do {
	  if (ioctl (cam_dev_fd, GIOC_CR_SET_SNAP, &one) < 0) {
	    if (errno != EINTR) {
	      if (errno == EBUSY) {
		break;
	      } else {
		if (errno == ENODATA) {
		  strcpy (windowname, "*** No Sync ***");
		  XStoreName (display, window, windowname);
		} else {
		  perror ("GIOC_CR_SET_SNAP");
		  Clean_Up ();
		}
	      }
	    }
	  } else {
	    break;
	  }
	} while (errno != EINTR);
	ICP_PRINTF ("-");
#if (0)
	if (Copy_Image() < 0) {
	  Clean_Up ();
	}
#endif
      }
      if (synchronization_mode == SELECT_MODE /* &&
	  acquisition_mode != SNAP_MODE */) {
	FD_ZERO (&wait_fdset);
	FD_SET (cam_dev_fd, &wait_fdset);
	if (snap_mode == TRIGGER_MODE) {
	  wait_time.tv_sec = 20;
	  wait_time.tv_usec = 0;
	} else {
	  wait_time.tv_sec = 1;
	  wait_time.tv_usec = 0;
	}
	error = select (getdtablesize(), &wait_fdset,
			NULL, NULL, &wait_time);
	switch (error) {
	case -1:
	  break;
	case 0:
	  strcpy (windowname, "*** No Sync ***");
	  XStoreName (display, window, windowname);
	  break;
	case 1:
	  if (FD_ISSET (cam_dev_fd, &wait_fdset)) {
	    ICP_PRINTF("/");
	  }
	  break;
	default:
	  perror ("unexpected return value");
	  Clean_Up ();
	}
	if (Copy_Image() < 0) {
	  Clean_Up ();
	}
      }
      if (synchronization_mode == POLL_MODE /* &&
	  acquisition_mode != SNAP_MODE */) {
	wait_fd.fd = cam_dev_fd;
	wait_fd.events = POLLIN | POLLERR;
	if (snap_mode == TRIGGER_MODE)
	  wait_timeout = 20000;	/* ms */
	else
	  wait_timeout = 1000;	/* ms */
	error = poll (&wait_fd, 1, wait_timeout);
	switch (error) {
	case -1:
	  break;
	case 0:
	  strcpy (windowname, "*** No Sync ***");
	  XStoreName (display, window, windowname);
	  break;
	case 1:
	  if (wait_fd.revents & POLLIN)
	    ICP_PRINTF("/");
	  break;
	default:
	  perror ("unexpected return value");
	  Clean_Up ();
	}
	if (Copy_Image() < 0) {
	  Clean_Up ();
	}
      }
      break;
    }

    if (operation_mode != READ_MODE && synchronization_mode == SIGNAL_MODE
	/* && acquisition_mode != SNAP_MODE */) {
      sigemptyset (&local_set);
      if (sigsuspend (&local_set) < 0)
	{
	  if (errno != EINTR)
	    {
	      perror ("sigsuspend error");
	      Clean_Up ();
	    }
	}
      if (timeout)
	{
	  strcpy (windowname, "*** No Sync ***");
	  XStoreName (display, window, windowname);
	  timeout = FALSE;
	}
    }

    /* Frame time stamp */
    
    if (operation_mode == BMA_MODE) {
      
      info_ptr = image_mmaped;
      
      image_info = (struct icp_info_t*)(info_ptr + image_width * image_height);
      
      memcpy (&frametime[0], &(image_info->time_stamp),
	      sizeof (struct timeval));
      
      xTime = frametime[0].tv_sec * 1000000 + frametime[0].tv_usec -
	starttime.tv_sec * 1000000 - starttime.tv_usec;
      sprintf (sTime, "%2ld:%2ld:%2ld", xTime / 60000000,
	       (xTime / 1000000) % 60,
	       (xTime / 1000) % 100);
    }
    
#if (0)
    /* save image */
    if (Save_Image (sTime, image_bmamap + bma_offset[bma_buffer],
		    bma_buffer) < 0) {
      Clean_Up ();
    }
#endif
    
    /* show image */
    if (shmmode) {
      XShmPutImage (display, window, context, image,
		    0, 0, 0, 0, image_width, image_height, 0);
      if (operation_mode == BMA_MODE) {
	XDrawImageString (display, window, context, 40, 30, sTime, 8);
      }
      XSync (display, TRUE);
    } else {
      XPutImage (display, window, context, image,
		 0, 0, 0, 0, image_width, image_height);
      if (operation_mode == BMA_MODE) {
	XDrawImageString (display, window, context, 40, 30, sTime, 8);
      }
    }
    
    ICP_PRINTF (".");
  }

  if (shmmode) {
    XShmDetach (display, &ShmInfo);
    shmdt (ShmInfo.shmaddr);
    shmctl (ShmInfo.shmid, IPC_RMID, 0);
    image->data = NULL;
  }
  if (image != NULL) {
    XDestroyImage (image);
  }
  if (context != NULL) {
    XFreeGC (display, context);
  }
  if (display != NULL) {
    XCloseDisplay (display);
  }

  if (cam_dev_fd != -1) {
    close (cam_dev_fd);
  }
  
  if (image_mmaped != NULL && mem_size != 0) {
    munmap (image_mmaped, mem_size);
  }
  
  ICP_PRINTF("\n");

  exit (EXIT_SUCCESS);
}

